#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.Contracts.ResponseObjects;
using HeuristicLab.Hive.Slave.Communication;
using HeuristicLab.Hive.Slave.Core.ConfigurationManager;

namespace HeuristicLab.Hive.Slave.Core.JobStorage {
  public class JobStorageManager {

    private static List<JobStorageInfo> storedJobsList = new List<JobStorageInfo>();

    private static String path = System.IO.Directory.GetCurrentDirectory() + "\\Hive.Slave.Jobs\\";

    public static void PersistObjectToDisc(String serverIP, long serverPort, Guid jobId, byte[] job) {
      String filename = serverIP + "." + serverPort + "." + jobId.ToString();
      JobStorageInfo info = new JobStorageInfo { JobID = jobId, ServerIP = serverIP, ServerPort = serverPort, TimeFinished = DateTime.Now };

      Stream jobstream = null;
      try {
        jobstream = File.Create(path + filename + ".dat");
        jobstream.Write(job, 0, job.Length);
        storedJobsList.Add(info);
        Logger.Info("Job " + info.JobID + " stored on the harddisc");
      }
      catch (Exception e) {
        Logger.Error("Exception: ", e);
      }
      finally {
        if (jobstream != null)
          jobstream.Close();
      }

      StoreJobList();
    }

    public static void CheckAndSubmitJobsFromDisc() {
      for (int index = storedJobsList.Count; index > 0; index--) {
        if (storedJobsList[index - 1].ServerIP == WcfService.Instance.ServerIp) {
          String filename = storedJobsList[index - 1].ServerIP + "." + storedJobsList[index - 1].ServerPort + "." + storedJobsList[index - 1].JobID.ToString();
          Logger.Info("Sending stored job " + storedJobsList[index - 1].JobID + " to the server");
          try {
            byte[] job = File.ReadAllBytes(path + filename + ".dat");
            if (WcfService.Instance.IsJobStillNeeded(storedJobsList[index - 1].JobID).StatusMessage == ResponseStatus.Ok) {
              ResponseResultReceived res = WcfService.Instance.StoreFinishedJobResultsSync(ConfigManager.Instance.GetClientInfo().Id, storedJobsList[index - 1].JobID, job, TimeSpan.Zero, null, true); // unfortunately we do not have the correct ExecutionTime since we would need to unzip the job for that
              Logger.Info("Sending of job " + storedJobsList[index - 1].JobID + " done");
            }
            SlaveStatusInfo.JobsProcessed++;
            storedJobsList.Remove(storedJobsList[index - 1]);
            File.Delete(path + filename + ".dat");
          }
          catch (Exception e) {
            Logger.Error("Job not on hdd but on list - deleting job from list ", e);
            storedJobsList.Remove(storedJobsList[index - 1]);
            StoreJobList();
          }
        }
      }
    }

    public static void StoreJobList() {
      XmlSerializer serializer = new XmlSerializer(typeof(List<JobStorageInfo>));
      TextWriter writer = new StreamWriter(Path.Combine(path, "list.xml"));
      serializer.Serialize(writer, storedJobsList);
      writer.Close();
    }

    static JobStorageManager() {
      Logger.Info("Restoring Joblist from Harddisk");
      if (!Directory.Exists(path))
        Directory.CreateDirectory(path);

      XmlSerializer serializer = new XmlSerializer(typeof(List<JobStorageInfo>));
      FileStream stream = null;
      if (File.Exists(Path.Combine(path, "list.xml"))) {
        try {
          stream = new FileStream(Path.Combine(path, "list.xml"), FileMode.Open);
          XmlTextReader reader = new XmlTextReader(stream);
          storedJobsList = (List<JobStorageInfo>)serializer.Deserialize(reader);
          Logger.Info("Loaded " + storedJobsList.Count + " Elements");
        }
        catch (Exception e) {
          Logger.Error("Exception while loading the Stored Job List", e);
        }
        finally {
          if (stream != null)
            stream.Dispose();
        }
      } else {
        Logger.Info("no stored jobs on harddisk, starting new list");
        storedJobsList = new List<JobStorageInfo>();
      }
    }
  }
}
